#include <stdio.h>
#include <math.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#define XSIZE 700
#define YSIZE 500

#define PI 3.1415926536

/* #define SLOW */

#ifdef SLOW
#define SPEEDKONST 0.08
#define THRUSTFORCE (0.6*SPEEDKONST)
#define TURNSPEED (0.35*SPEEDKONST)
#define GRAVITATION (0.07*SPEEDKONST)
#else
#define SPEEDKONST 0.5
#define THRUSTFORCE (0.6*SPEEDKONST)
#define TURNSPEED (0.25*SPEEDKONST)
#define GRAVITATION (0.07*SPEEDKONST)
#endif


#define SHIPMASS 1.0
#define BALLMASS 2.0
#define LINELENGTH 80.0
#define BALLRADIUS 9
#define GOALRADIUS 13

/* 1.0 och 0.4 lmpligt fr stel pinne. */
#define LINEK 1.0 /* Fjderkonstant */
#define LINED 0.4 /* Fjderdmpning */


#define FJADER /* Bestmmer om "snret" skall vara en fjder
		  eller en gumminsnodd */

#define MAXWORLDSIZE 100
#define MAXPOINTS 100

typedef struct {
  double arg;
  double len;
} vectorT;


typedef struct {
  long x;
  long y;
} punktT;

typedef struct {
  int size;
  long minx, miny, maxx, maxy;
  punktT *punkter;
} blockT;

/* Fltet head innehller fr skeppet "gasvektorn". Fr bollen innehller
   den "snrvektorn" frn skeppet till bollen. */
typedef struct {
  vectorT head;
  vectorT vel;
} objectT;

static Display *g_disp;
static Window   g_window;
static Pixmap g_pixmap = (Pixmap)NULL;
static GC       g_gc;
static int      g_screen;
static int      g_screennum;

static int g_signal = 0;

static int g_shipdef[10][2] = {
  13, 0,
  -3, -10,
  -6, -10,
  -10, -6,
  -6, -3,
  -6, 3,
  -10, 6,
  -6, 10,
  -3, 10,
  13, 0
};

#define SHIPRAD 14 /* Radie p skeppet */

#define MIN(a,b) (( (a) > (b) ) ? (b) : (a))
#define MAX(a,b) (( (a) > (b) ) ? (a) : (b))

static long g_startx = 0;
static long g_starty = 0;
static long g_bstartx = 0;
static long g_bstarty = 0;
static long g_goalx = 0;
static long g_goaly = 100;
static int g_level = 1;

static void sighandler(int s);
static inline void printvect(vectorT *v);
static inline void torect(vectorT *v, double *x, double *y);
static inline void vectoradd(vectorT *v1, vectorT *v2, vectorT *result);
static inline void vectorsub(vectorT *v1, vectorT *v2, vectorT *result);
static inline void topol(double x, double y, vectorT *v);
static void calculate_velocities(objectT *ship, objectT *ball,
				 double *x, double *y,
				 double *bx, double *by, int ball_loose,
				 int hasball, int catching, vectorT *grav);

static int init_world(blockT *world, int *worldsize);
static void make_world(blockT *world, int *worldsize);
static void init_disp(void);
static void blackpen(void);
static void whitepen(void);
static void drawobjs(objectT *ship, objectT *ball, double x, double y,
		     double bx, double by, int hasball, int catching,
		     long offsetx, long offsety, Drawable dest);
static void drawworld(blockT *world, int worldsize,
		      long offsetx, long offsety, Drawable dest);

static inline int colcheck(int xa1, int xa2, int ya1, int ya2,
			   int xb1, int xb2, int yb1, int yb2);


int colcheck(int xa1, int xa2, int ya1, int ya2,
	     int xb1, int xb2, int yb1, int yb2) {

  double r, s;
  int dom;

  /* Box-check */
  if(xb2>xa2 && xb2>xa1 && xb1>xa2 && xb1>xa1) return 0;
  if(xb2<xa2 && xb2<xa1 && xb1<xa2 && xb1<xa1) return 0;
  if(yb2>ya2 && yb2>ya1 && yb1>ya2 && yb1>ya1) return 0;
  if(yb2<ya2 && yb2<ya1 && yb1<ya2 && yb1<ya1) return 0;

  dom = (xa2-xa1)*(yb2-yb1) - (ya2-ya1)*(xb2-xb1);
  if(dom == 0) return 0;

  r = ((ya1-yb1)*(xb2-xb1) - (xa1-xb1)*(yb2-yb1)) / (double)dom;
  if(r < 0.0 || r > 1.0) return 0;

  s = ((ya1-yb1)*(xa2-xa1) - (xa1-xb1)*(ya2-ya1)) / (double)dom;
  if(s < 0.0 || s > 1.0) return 0;
  
  return 1;
}


void drawobjs(objectT *ship, objectT *ball, double x, double y,
	      double bx, double by, int hasball, int catching,
	      long offsetx, long offsety, Drawable dest) {
  XPoint pts1[10];
  double cosv, sinv, shipx, shipy, gshipx, gshipy, gballx, gbally;
  int i, j;
  static int step = 0;
  int xa1, xa2, ya1, ya2, xb1, xb2, yb1, yb2;
  double dist, l, s, r;

  if(++step >= 6) step = 0;
  
  cosv = cos(ship->head.arg);
  sinv = sin(ship->head.arg);

  gshipx = x;
  gshipy = y;

  for(i=0; i<10; i++) {
    shipx = g_shipdef[i][0];
    shipy = g_shipdef[i][1];
    pts1[i].x =  (cosv * shipx - sinv * shipy) + gshipx + offsetx;
    pts1[i].y =  YSIZE-((sinv * shipx + cosv * shipy) + gshipy + offsety);
  }

  XDrawLines(g_disp, dest, g_gc, pts1, 10, CoordModeOrigin);

  if(!hasball) {
    gballx = bx;
    gbally = by;
  }
  else {
    torect(&ball->head, &gballx, &gbally);
    gballx += gshipx;
    gbally += gshipy;
  }

  if(!hasball && catching && step < 3) {
    XDrawLine(g_disp, dest, g_gc, gshipx + offsetx, YSIZE-gshipy - offsety,
	      gballx + offsetx, YSIZE-gbally - offsety);
    
  }

  if(hasball) {
#ifndef FJADER
    if(ball->head.len < LINELENGTH) 
      XSetLineAttributes(g_disp, g_gc,
			 0, LineOnOffDash, CapProjecting, JoinBevel);
#endif  
    XDrawLine(g_disp, dest, g_gc, gshipx + offsetx, YSIZE-gshipy - offsety,
	      gballx + offsetx, YSIZE-gbally - offsety);
    
#ifndef FJADER
    if(ball->head.len < LINELENGTH) 
      XSetLineAttributes(g_disp, g_gc,
			 0, LineSolid, CapProjecting, JoinBevel);
#endif
  }

  XDrawArc(g_disp, dest, g_gc, gballx-BALLRADIUS + offsetx,
	   YSIZE-gbally-BALLRADIUS - offsety,
	   BALLRADIUS*2, BALLRADIUS*2, 0, 360*64);

  XDrawArc(g_disp, dest, g_gc, g_goalx-GOALRADIUS + offsetx,
	   YSIZE-g_goaly-GOALRADIUS - offsety,
	   GOALRADIUS*2, GOALRADIUS*2, 0, 360*64);
}

void blackpen(void) {
  XSetForeground(g_disp, g_gc, BlackPixel(g_disp, g_screennum));
}


void whitepen(void) {
  XSetForeground(g_disp, g_gc, WhitePixel(g_disp, g_screennum));
}


void init_disp(void) {
  XWindowAttributes xwa;

  g_disp = XOpenDisplay(0);
  if(!g_disp) exit(1);
  g_screennum = DefaultScreen(g_disp);
  g_window = XCreateSimpleWindow(g_disp,
			      RootWindow(g_disp, g_screennum),
			      0, 0,
			      XSIZE,YSIZE,
			      0,WhitePixel(g_disp, g_screennum),
			      BlackPixel(g_disp,g_screennum));
  g_gc = XDefaultGC(g_disp,g_screennum);

  XSelectInput(g_disp, g_window, KeyPressMask | KeyReleaseMask);
  XSetGraphicsExposures(g_disp, g_gc, 0);
  
  XMapWindow(g_disp,g_window);

  XSetBackground(g_disp, g_gc, BlackPixel(g_disp, g_screennum));

  XSync(g_disp,0);

  while(XGetWindowAttributes(g_disp, g_window, &xwa) &&
	xwa.map_state != IsViewable) usleep(100000);

  g_pixmap = XCreatePixmap(g_disp,
			   RootWindow(g_disp, g_screennum),
			   XSIZE, YSIZE,
			   DefaultDepth(g_disp, g_screennum));


  blackpen();
  XFillRectangle(g_disp, g_pixmap, g_gc,
		 0, 0, XSIZE, YSIZE);
}


void calculate_velocities(objectT *ship, objectT *ball, double *x, double *y,
			  double *bx, double *by, int ball_loose,
			  int hasball, int catching, vectorT *grav) {
  vectorT stringf, vrel, ballf, shipf, gravf;
  double damping, xx, yy;

  if(hasball) {
    /* Berkna gummibandets spnning */
    stringf.arg = ball->head.arg; /* Kraften riktad lngs snret */
    stringf.len = (ball->head.len - LINELENGTH) * LINEK;
    
#ifndef FJADER
    if(stringf.len < 0.0) stringf.len = 0.0;
#endif
    
    /* Kulans hastighet relativt skeppet */
    vectorsub(&ball->vel, &ship->vel, &vrel);
    
    /* Relativ hastighet lngs snret gnger dmpningsfaktor = dmpning.
       Kraften lngs gummibandet minskar eller kar med dmpningen. */
    damping = vrel.len * cos(stringf.arg - vrel.arg) * LINED;
    
#ifndef FJADER
    if(ball->head.len < LINELENGTH) damping = 0.0;
#endif
    
    stringf.len += damping;
    
#ifndef FJADER
    /* Dmpningen om kulan och skeppet rr sig mot varandra kan aldrig
       vara s stor att spnningen blir mindre n 0, dvs att gummibandet
       "skjuter p". */
    if(stringf.len < 0.0) stringf.len = 0.0;
#endif
    
  }
  
  /* Berkna resulterande kraft p skeppet */
  gravf.arg = grav->arg;
  gravf.len = grav->len * SHIPMASS;
  vectoradd(&ship->head, &gravf, &shipf);
  if(hasball) vectoradd(&stringf, &shipf, &shipf);
  shipf.len /= SHIPMASS;
    
  /* Berkna resulterande kraft p kulan */
  gravf.len = grav->len * BALLMASS;
  if(hasball) vectorsub(&gravf, &stringf, &ballf);
  if(ball_loose) {
    ballf.arg = gravf.arg;
    ballf.len = gravf.len;
  }
  
  /* Dividera med massan fr att f acceleration */
  if(hasball || ball_loose) ballf.len /= BALLMASS;

  /* ndra hastigheter */
  vectoradd(&shipf, &ship->vel, &ship->vel);
  if(hasball || ball_loose) vectoradd(&ballf, &ball->vel, &ball->vel);

  /* Uppdatera positioner */
  if(hasball || ball_loose) {
    vectoradd(&ball->vel, &ball->head, &ball->head);
    if(hasball) vectorsub(&ball->head, &ship->vel, &ball->head);
  }

  torect(&ship->vel, &xx, &yy);
  *x += xx;
  *y += yy;

  if(hasball) {
    torect(&ball->head, bx, by);
    *bx += *x;
    *by += *y;
  }

  if(ball_loose) {
    torect(&ball->vel, &xx, &yy);
    *bx += xx;
    *by += yy;
  }
}


void printvect(vectorT *v) {
  printf("arg: %f  len: %f\n", v->arg * 180 / PI, v->len);
}


void torect(vectorT *v, double *x, double *y) {
  *y = v->len * sin(v->arg);
  *x = v->len * cos(v->arg);
}

void vectoradd(vectorT *v1, vectorT *v2, vectorT *result) {
  double x1, y1, x2, y2;
  torect(v1, &x1, &y1);
  torect(v2, &x2, &y2);
  x1 += x2;
  y1 += y2;
  topol(x1, y1, result);
}


void vectorsub(vectorT *v1, vectorT *v2, vectorT *result) {
  double x1, y1, x2, y2;
  torect(v1, &x1, &y1);
  torect(v2, &x2, &y2);
  x1 -= x2;
  y1 -= y2;
  topol(x1, y1, result);
}


void topol(double x, double y, vectorT *v) {
  v->len = sqrt(x*x + y*y);
  v->arg = atan2(y, x);
}



int main(int argc, char **argv) {
  int left=0, right=0, thrust=0, space=0, lockspace = 0;
  int i, j, k, worldsize = 0, hasball = 0, catching = 0, coll = 0;
  int ball_loose=0, minx, maxx, miny, maxy;
  KeyCode leftKey, rightKey, thrustKey, spaceKey, makeKey, key;
  objectT ball, ship;
  vectorT grav;
  double x, y; /* Skeppets position */
  double bx, by; /* Bollens position */
  double xx, yy, cosv, sinv, dist, l, s, r, xb1, xb2, yb1, yb2;
  XEvent report;
  blockT world[MAXWORLDSIZE];
  double offsetx = 0, offsety = 0, offsetvx = 0, offsetvy = 0;
  punktT pts[10], *wpts;

  if(argc ==2) {
    if(sscanf(argv[1], "%d", &g_level) != 1) g_level = 1;
  }
  
  grav.arg = 3*PI/2;
  grav.len = GRAVITATION;

  ship.head.arg = PI/2;
  ship.head.len = 0;
  ship.vel.arg = 0;
  ship.vel.len = 0;

  ball.head.arg = -PI/2;
  ball.head.len = LINELENGTH;
  ball.vel.arg = 0;
  ball.vel.len = 0;

  

  init_disp();
  leftKey = XKeysymToKeycode(g_disp, XK_A);
  rightKey = XKeysymToKeycode(g_disp, XK_S);
  thrustKey = XKeysymToKeycode(g_disp, XK_Shift_R);
  spaceKey = XKeysymToKeycode(g_disp, XK_space);
  makeKey = XKeysymToKeycode(g_disp, XK_M);
  
  init_world(world, &worldsize);

  x = g_startx;
  y = g_starty;
  bx = g_bstartx;
  by = g_bstarty;

  signal(SIGHUP, sighandler);
  signal(SIGINT, sighandler);
  XAutoRepeatOff(g_disp);

  
  while(1) {

    if(g_signal) {
      XAutoRepeatOn(g_disp);
      XSync(g_disp, 0);
      exit(0);
    }

    whitepen();
    drawobjs(&ship, &ball, x, y, bx, by, hasball, catching,
	     offsetx, offsety, g_pixmap);
    drawworld(world, worldsize, offsetx, offsety, g_pixmap);

    XCopyArea(g_disp, g_pixmap, g_window, g_gc,
	      0, 0, XSIZE, YSIZE, 0, 0);

    blackpen();
    XFillRectangle(g_disp, g_pixmap, g_gc,
		   0, 0, XSIZE, YSIZE);
    /* drawobjs(&ship, &ball, x, y, g_pixmap); */
    
    XSync(g_disp, 0);

    /* Ml? */
    
    if(hasball) {
      torect(&ball.head, &xx, &yy);
      xx += x;
      yy += y;
    }
    if(ball_loose) {
      xx = bx;
      yy = by;
    }

    if(sqrt((g_goalx-xx)*(g_goalx-xx) + (g_goaly-yy)*(g_goaly-yy)) <
       BALLRADIUS+GOALRADIUS) {
      printf("Level %d complete!\n", g_level);
      g_level++;

      sleep(1);

      if(init_world(world, &worldsize) < 0) {
	XAutoRepeatOn(g_disp);
	XSync(g_disp, 0);
	exit(0);
      }


      ship.head.arg = PI/2;
      ship.head.len = 0;
      ship.vel.arg = 0;
      ship.vel.len = 0;
      
      ball.head.arg = -PI/2;
      ball.head.len = LINELENGTH;
      ball.vel.arg = 0;
      ball.vel.len = 0;
      x = g_startx;
      y = g_starty;
      bx = g_bstartx;
      by = g_bstarty;
      hasball = 0;
      catching = 0;
      ball_loose = 0;
      coll = 0;
      continue;
    }

    /* Kollisionsdetektering */
    for(i=0; i<worldsize; i++) {

      /* Skepp */
      if(x+SHIPRAD >= world[i].minx && x-SHIPRAD <= world[i].maxx &&
	 y+SHIPRAD >= world[i].miny && y-SHIPRAD <= world[i].maxy) {

	/* Vi blir tvungna att kolla varje linje */

	/* Berkna skeppets position */
	cosv = cos(ship.head.arg);
	sinv = sin(ship.head.arg);
	for(k=0; k<10; k++) {
	  xx = g_shipdef[k][0];
	  yy = g_shipdef[k][1];
	  pts[k].x = (cosv * xx - sinv * yy) + x;
	  pts[k].y = (sinv * xx + cosv * yy) + y;
	}
	
	wpts = world[i].punkter;
	
	for(j=0; j<world[i].size-1; j++) {
	  for(k=0; k<9; k++) {
	    if(colcheck(pts[k].x, pts[k+1].x, pts[k].y, pts[k+1].y,
		wpts[j].x, wpts[j+1].x, wpts[j].y, wpts[j+1].y)) coll = 1;
	  }
	}
      }
      
      /* Skepp kollat. Kolla snre och boll. */
      if(hasball || ball_loose) {
	if(hasball) {
	  torect(&ball.head, &xx, &yy);
	  xx += x;
	  yy += y;
	  
	  /* Snre */

	  minx = MIN(x, xx);
	  maxx = MAX(x, xx);
	  miny = MIN(y, yy);
	  maxy = MAX(y, yy);


	  if(!(maxx < world[i].minx || minx > world[i].maxx ||
	     maxy < world[i].miny || miny > world[i].maxy)) {
	    /* Kolla varje linje */
	    wpts = world[i].punkter;
	    for(j=0; j<world[i].size-1; j++) {
	      if(colcheck(x, xx, y, yy,
		  wpts[j].x, wpts[j+1].x, wpts[j].y, wpts[j+1].y)) coll = 1;
	    }
	  }
	}

	if(ball_loose) {
	  xx = bx;
	  yy = by;
	}
	
	/* Boll */
	if(xx+BALLRADIUS >= world[i].minx && xx-BALLRADIUS <= world[i].maxx &&
	   yy+BALLRADIUS >= world[i].miny && yy-BALLRADIUS <= world[i].maxy) {

	  /* Kolla varje linje */
	  wpts = world[i].punkter;
	  for(j=0; j<world[i].size-1; j++) {
	    xb1 = wpts[j].x;
	    xb2 = wpts[j+1].x;
	    yb1 = wpts[j].y;
	    yb2 = wpts[j+1].y;
	    l = sqrt((xb2-xb1)*(xb2-xb1) + (yb2-yb1)*(yb2-yb1));
	    s = ((yb1-yy)*(xb2-xb1) - (xb1-xx)*(yb2-yb1)) / (l*l);
	    r = ((yb1-yy)*(yb1-yb2) - (xb1-xx)*(xb2-xb1)) / (l*l);
	    dist = fabs(s*l);
	    if(r>=0 && r<=1 && dist<BALLRADIUS) coll = 1;
	    
	    /* Kolla avstnd till ndpunkterna */
	    dist = sqrt((xb1-xx)*(xb1-xx) + (yb1-yy)*(yb1-yy));
	    if(dist < BALLRADIUS) coll = 1;
	    dist = sqrt((xb2-xx)*(xb2-xx) + (yb2-yy)*(yb2-yy));
	    if(dist < BALLRADIUS) coll = 1;
	  }
	}
      }
    }

    if(coll) {
      sleep(1);
      ship.head.arg = PI/2;
      ship.head.len = 0;
      ship.vel.arg = 0;
      ship.vel.len = 0;
      
      ball.head.arg = -PI/2;
      ball.head.len = LINELENGTH;
      ball.vel.arg = 0;
      ball.vel.len = 0;
      x = g_startx;
      y = g_starty;
      bx = g_bstartx;
      by = g_bstarty;
      hasball = 0;
      catching = 0;
      ball_loose = 0;
      coll = 0;
      continue;
    }
    
    if(XPending(g_disp)) {
      XNextEvent(g_disp, &report);
      switch(report.type) {
      case KeyPress:
	key = report.xkey.keycode;
	if(key == leftKey) left = 1;
	else if(key == rightKey) right = 1;
	else if(key == thrustKey) thrust = 1;
	else if(key == spaceKey) space = 1;
	else if(key == makeKey) {
	  make_world(world, &worldsize);
	  ship.head.arg = PI/2;
	  ship.head.len = 0;
	  ship.vel.arg = 0;
	  ship.vel.len = 0;
	  
	  ball.head.arg = -PI/2;
	  ball.head.len = LINELENGTH;
	  ball.vel.arg = 0;
	  ball.vel.len = 0;
	  
	  x = g_startx;
	  y = g_starty;
	  bx = g_bstartx;
	  by = g_bstarty;
	  hasball = 0;
	  ball_loose = 0;
	  coll = 0;
	}
	break;
      case KeyRelease:
	key = report.xkey.keycode;
	if(key == leftKey) left = 0;
	else if(key == rightKey) right = 0;
	else if(key == thrustKey) thrust = 0;
	else if(key == spaceKey) space = 0;
	break;
      default:
	break;
      }
    }

    if(right) ship.head.arg -= TURNSPEED;
    if(left) ship.head.arg += TURNSPEED;
    if(thrust) ship.head.len = THRUSTFORCE;
    else ship.head.len = 0.0;

    if(!space) {
      catching = 0;
      lockspace = 0;
    }

    if(catching == 0 && hasball && space && !lockspace) {
      /* Slpp bollen */
      hasball = 0;
      lockspace = 1;
      ball_loose = 1;
    }

    if(!hasball && !catching && space && !lockspace) {
      if(sqrt((bx-x)*(bx-x) + (by-y)*(by-y)) < LINELENGTH) {
	catching = 1;
	lockspace = 1;
      }
    }

    if(catching) {
      if(sqrt((bx-x)*(bx-x) + (by-y)*(by-y)) > LINELENGTH) {
	catching = 0;
	ship.vel.arg = 0;
	ship.vel.len = 0;
	ball.vel.arg = 0;
	ball.vel.len = 0;
	ball.head.len = LINELENGTH;
	ball.head.arg = atan2(by-y, bx-x);
	catching = 0;
	hasball = 1;
	ball_loose = 0;
      }
    }

    if(ship.head.arg >= 2*PI) ship.head.arg -= 2*PI;

    calculate_velocities(&ship, &ball, &x, &y,
			 &bx, &by, ball_loose, hasball, catching, &grav);

#ifdef SLOW
    dist = (XSIZE / 2 - x - offsetx) / 400.0;
#else
    dist = (XSIZE / 2 - x - offsetx) / 200.0;
#endif
    if(dist < 0) dist = -dist*dist;
    else dist = dist*dist;
    offsetvx += dist;
    offsetvx -= offsetvx / 15.0;

#ifdef SLOW
    dist = (YSIZE / 2 - y - offsety) / 400.0;
#else
    dist = (YSIZE / 2 - y - offsety) / 200.0;
#endif
    if(dist < 0) dist = -dist*dist;
    else dist = dist*dist;
    offsetvy += dist;
    offsetvy -= offsetvy / 15.0;
    
    offsetx += offsetvx;
    offsety += offsetvy;

    if(offsetx+x < LINELENGTH+10) offsetx = LINELENGTH+10-x;
    if(offsetx+x > XSIZE-(LINELENGTH+10)) offsetx = XSIZE-(LINELENGTH+10)-x;
    if(offsety+y < LINELENGTH+10) offsety = LINELENGTH+10-y;
    if(offsety+y > YSIZE-(LINELENGTH+10)) offsety = YSIZE-(LINELENGTH+10)-y;
  }
}


void sighandler(int s) {
  g_signal = 1;
}


int init_world(blockT *world, int *worldsize) {
  punktT *pts;
  int i, j;
  FILE *fil;
  char buf[100], buf2[20];

  *worldsize = 0;
  strcpy(buf, "thrust.map");
  sprintf(buf2, "%d", g_level);
  strcat(buf, buf2);
  
  fil = fopen(buf, "r");
  if(!fil) {
    printf("Could not open file for level %d: %s\n", g_level, buf);
    return -1;
  }
  
  if(fscanf(fil, "%d", worldsize) != 1) {
    fclose(fil);
    printf("Read error!\n");
    *worldsize = 0;
    return -1;
  }

  for(i=0; i<*worldsize; i++) {
    if(fscanf(fil, "%d %ld %ld %ld %ld\n",
	      &world[i].size, &world[i].minx, &world[i].miny,
	      &world[i].maxx, &world[i].maxy) != 5) {
      fclose(fil);
      printf("Read error!\n");
      *worldsize = 0;
      return -1;
    }

    pts = (punktT *)malloc(sizeof(punktT) * world[i].size);
    if(!pts) {
      printf("Slut p minne!\n");
      fclose(fil);
      *worldsize = 0;
      return -1;
    }

    for(j=0; j<world[i].size; j++) {
      if(fscanf(fil, "%ld %ld", &pts[j].x, &pts[j].y) != 2) {
      fclose(fil);
      printf("Read error!\n");
      *worldsize = 0;
      return -1;
      }
    }

    world[i].punkter = pts;
  }

  if(fscanf(fil, "%ld %ld %ld %ld", &g_startx, &g_starty, &g_bstartx,
	    &g_bstarty) != 4) {
    fclose(fil);
    printf("Read error!\n");
    *worldsize = 0;
    return -1;
  }

  if(fscanf(fil, "%ld %ld", &g_goalx, &g_goaly) != 2) {
    fclose(fil);
    printf("Read error!\n");
    *worldsize = 0;
    return -1;
  }

  return 0;
}


void drawworld(blockT *world, int worldsize,
	       long offsetx, long offsety, Drawable dest) {
  static XPoint pts[MAXPOINTS];
  int i, j;

  for(i=0; i<worldsize; i++) {

    if(world[i].minx + offsetx > XSIZE ||
       world[i].maxx + offsetx < 0 ||
       YSIZE - world[i].miny - offsety < 0 ||
       YSIZE - world[i].maxy - offsety > YSIZE) {
      continue;
    }
    
    for(j=0; j<world[i].size; j++) {
      pts[j].x = world[i].punkter[j].x + offsetx;
      pts[j].y = YSIZE - world[i].punkter[j].y - offsety;
    }
    XFillPolygon(g_disp, dest, g_gc, pts, world[i].size,
		 Nonconvex, CoordModeOrigin);
  }
}


void make_world(blockT *world, int *worldsize) {
  XEvent report;
  KeyCode key;
  long ptrx = 0, ptry = 0, offsetx = 0, offsety = 0, diffx, diffy;
  long bestx, besty, minx, maxx, miny, maxy;
  double bestdist, dst;
  static XPoint pts[MAXPOINTS];
  punktT *pts2;
  int nrofpts, drawing = 0, i, dum, xx, yy, j, bestblock;
  Window dumwin;
  KeyCode lockKey, deleteKey, exitKey, saveKey, ballKey, positionKey;
  KeyCode goalKey;
  int lock=0;
  objectT ball, ship;
  FILE *fil;

  ship.head.arg = PI/2;
  ship.head.len = 0;
  ship.vel.arg = 0;
  ship.vel.len = 0;

  ball.head.arg = -PI/2;
  ball.head.len = LINELENGTH;
  ball.vel.arg = 0;
  ball.vel.len = 0;

  lockKey = XKeysymToKeycode(g_disp, XK_Shift_L);
  deleteKey = XKeysymToKeycode(g_disp, XK_D);
  exitKey = XKeysymToKeycode(g_disp, XK_X);
  saveKey = XKeysymToKeycode(g_disp, XK_S);
  ballKey = XKeysymToKeycode(g_disp, XK_B);
  positionKey = XKeysymToKeycode(g_disp, XK_P);
  goalKey = XKeysymToKeycode(g_disp, XK_G);
  
  XSelectInput(g_disp, g_window, KeyPressMask | KeyReleaseMask |
	       PointerMotionMask | ButtonPressMask | ButtonReleaseMask);

  XQueryPointer(g_disp, g_window, &dumwin, &dumwin, &dum, &dum,
		&xx, &yy, &dum);
  ptrx = xx - offsetx;
  ptry = YSIZE - yy - offsety;
  
  while(1) {
    if(g_signal) {
      XAutoRepeatOn(g_disp);
      XSync(g_disp, 0);
      exit(0);
    }
  
    whitepen();
    drawobjs(&ship, &ball, g_startx, g_starty, g_bstartx, g_bstarty,
	     0, 0, offsetx, offsety, g_pixmap);
    drawworld(world, *worldsize, offsetx, offsety, g_pixmap);

    XDrawLine(g_disp, g_pixmap, g_gc, ptrx + offsetx - 5, YSIZE-ptry-offsety,
	      ptrx + offsetx + 5, YSIZE-ptry - offsety);
    XDrawLine(g_disp, g_pixmap, g_gc, ptrx + offsetx, YSIZE-ptry-offsety - 5,
	      ptrx + offsetx, YSIZE-ptry - offsety + 5);

    if(drawing) {
      XDrawLines(g_disp, g_pixmap, g_gc, pts, nrofpts, CoordModeOrigin);
      XDrawLine(g_disp, g_pixmap, g_gc, pts[nrofpts-1].x, pts[nrofpts-1].y,
		ptrx + offsetx, YSIZE-ptry - offsety);
    }

    XCopyArea(g_disp, g_pixmap, g_window, g_gc,
	      0, 0, XSIZE, YSIZE, 0, 0);
    blackpen();
    XFillRectangle(g_disp, g_pixmap, g_gc,
		   0, 0, XSIZE, YSIZE);
    
    XSync(g_disp, 0);

    diffx = 0;
    diffy = 0;
    
    if((ptrx+offsetx) < XSIZE / 5) {
      diffx = 20 * (XSIZE/5 - (ptrx+offsetx)) / (XSIZE/5);
      offsetx +=  diffx;
      ptrx -= diffx;
    }
    else if((ptrx+offsetx) > (XSIZE - XSIZE / 5)) {
      diffx = 20 * (XSIZE - XSIZE/5 - (ptrx+offsetx)) / (XSIZE/5);
      offsetx += diffx;
      ptrx -=  diffx;
    }

    if((ptry+offsety) < YSIZE / 5) {
      diffy = 20 * (YSIZE/5 - (ptry+offsety)) / (YSIZE/5);
      offsety +=  diffy;
      ptry -= diffy;
    }
    else if((ptry+offsety) > (YSIZE - YSIZE / 5)) {
      diffy = 20 * (YSIZE - YSIZE/5 - (ptry+offsety)) / (YSIZE/5);
      offsety += diffy;
      ptry -=  diffy;
    }
    
    if(drawing && (diffx || diffy)) {
      for(i=0; i<nrofpts; i++) {
	pts[i].x += diffx;
	pts[i].y -= diffy;
      }
    }
    
    if(XPending(g_disp)) {
      XNextEvent(g_disp, &report);
      switch(report.type) {
      case KeyPress:
	key = report.xkey.keycode;
	if(key == lockKey) lock=1;
	else if(key == exitKey) return;
	else if(key == ballKey) {
	  g_bstartx = ptrx;
	  g_bstarty = ptry;
	}
	else if(key == goalKey) {
	  g_goalx = ptrx;
	  g_goaly = ptry;
	}
	else if(key == positionKey) {
	  g_startx = ptrx;
	  g_starty = ptry;
	}
	else if(key == saveKey) {
	  fil = fopen("thrust.map", "r");
	  if(fil) {
	    printf("Filen thrust.map existerar redan!\n");
	    fclose(fil);
	    break;
	  }
	  fil = fopen("thrust.map", "w");
	  if(!fil) {
	    printf("Kunde inte ppna thrust.map fr skrivning!\n");
	    break;
	  }
	  fprintf(fil, "%d\n", *worldsize);
	  for(i=0; i<*worldsize; i++) {
	    fprintf(fil, "%d %ld %ld %ld %ld\n", world[i].size,
		   world[i].minx, world[i].miny,
		   world[i].maxx,world[i].maxy);
	    pts2 = world[i].punkter;
	    for(j=0; j<world[i].size; j++) {
	      fprintf(fil, "%ld %ld\n", pts2[j].x, pts2[j].y);
	    }
	  }
	  fprintf(fil, "%ld %ld %ld %ld\n", g_startx, g_starty,
		  g_bstartx, g_bstarty);
	  fprintf(fil, "%ld %ld\n", g_goalx, g_goaly);
	  fclose(fil);
	}
	else if(key == deleteKey && *worldsize > 0) {
	  bestdist = 9E20;
	  for(i=0; i<*worldsize; i++) {
	    pts2 = world[i].punkter;
	    for(j=0; j<world[i].size; j++) {
	      diffx = pts2[j].x - ptrx;
	      diffy = pts2[j].y - ptry;
	      dst = sqrt((double)(diffx*diffx + diffy*diffy));
	      if(dst < bestdist) {
		bestdist = dst;
		bestblock = i;
	      }
	    }
	  }
	  free(world[bestblock].punkter);
	  if(bestblock == *worldsize - 1) (*worldsize)--;
	  else {
	    for(i=bestblock+1; i<*worldsize; i++) {
	      world[i-1].size = world[i].size;
	      world[i-1].minx = world[i].minx;
	      world[i-1].miny = world[i].miny;
	      world[i-1].maxx = world[i].maxx;
	      world[i-1].maxy = world[i].maxy;
	      world[i-1].punkter = world[i].punkter;
	    }
	    (*worldsize)--;
	  }
	}
	break;
      case KeyRelease:
	key = report.xkey.keycode;
	if(key == lockKey) lock=0;
	break;
      case ButtonPress:
	if(report.xbutton.button == 1) {
	  if(!drawing) nrofpts = 0;
	  drawing = 1;
	  if(nrofpts >= MAXPOINTS-1) {
	    printf("Object too big!\n");
	    drawing = 0;
	    break;
	  }
	  pts[nrofpts].x = ptrx + offsetx;
	  pts[nrofpts].y = YSIZE - ptry - offsety;
	  nrofpts++;
	}
	else if(report.xbutton.button == 2 && drawing) {
	  drawing = 0;
	  nrofpts++;
	  if(nrofpts < 4) {
	    printf("Fr f punkter!\n");
	    break;
	  }
	  pts2 = (punktT *)malloc(sizeof(punktT) * nrofpts);

	  if(!pts2) {
	    printf("Out of memory!\n");
	    XAutoRepeatOn(g_disp);
	    XSync(g_disp, 0);
	    exit(0);
	  }

	  minx = 999999999;
	  miny = 999999999;
	  maxx = -999999999;
	  maxy = -999999999;

	  for(i=0; i<nrofpts-1; i++) {
	    pts2[i].x = pts[i].x - offsetx;
	    pts2[i].y = YSIZE - pts[i].y - offsety;
	    if(pts2[i].x > maxx) maxx = pts2[i].x;
	    if(pts2[i].y > maxy) maxy = pts2[i].y;
	    if(pts2[i].x < minx) minx = pts2[i].x;
	    if(pts2[i].y < miny) miny = pts2[i].y;
	  }
	  pts2[nrofpts-1].x = pts2[0].x;
	  pts2[nrofpts-1].y = pts2[0].y;

	  world[*worldsize].punkter = pts2;
	  world[*worldsize].size = nrofpts;
	  world[*worldsize].minx = minx;
	  world[*worldsize].miny = miny;
	  world[*worldsize].maxx = maxx;
	  world[*worldsize].maxy = maxy;
	  
	  (*worldsize)++;
	  
	}
	else drawing = 0;
	
	break;
      case ButtonRelease:
	break;
      case MotionNotify:

	ptrx = report.xmotion.x - offsetx;
	ptry = YSIZE - report.xmotion.y - offsety;

	if(lock && *worldsize > 0) {
	  bestdist = 9E20;
	  for(i=0; i<*worldsize; i++) {
	    pts2 = world[i].punkter;
	    for(j=0; j<world[i].size; j++) {
	      diffx = pts2[j].x - ptrx;
	      diffy = pts2[j].y - ptry;
	      dst = sqrt((double)(diffx*diffx + diffy*diffy));
	      if(dst < bestdist) {
		bestdist = dst;
		bestx = pts2[j].x;
		besty = pts2[j].y;
	      }
	    }
	  }
	  ptrx = bestx;
	  ptry = besty;
	}
	break;
      default:
	break;
      }
    }
  }  
}
